<?php
/*
    author: Matheus P. Alves Nogueira
    date: 2017
    e-mail: matheus.nogueira@teknisa.com
*/

namespace Teknisa\MAN\Services;

use Doctrine\ORM;
use Teknisa\MAN\Utilities;
use Zeedhi\Framework\DataSource\FilterCriteria;
use Zeedhi\Framework\DB\StoredProcedure\StoredProcedure;
use Zeedhi\Framework\DB\StoredProcedure\Param;
use Zeedhi\Framework\DTO\Request;
use \Setforms2\Shared\ORM\DateTime;

class ConsolidacaoVendasService {

    // status 'SUCCESS' para consolidação concluída 
    const SUCCESS_C = 'SUCCESS';
    // status 'FAIL' para consolidação cancelada
    const FAIL_C = 'FAIL';
    // status 'FAIL_DATE' para consolidação cancelada com todas as vendas da data irregulares 
    const FAIL_DATE_C = 'FAIL_DATE';

    protected $entityManager;
    protected $environment;
    protected $dataSourceNames = Array(
        'filterConsolidacao' => 'vnd05400_consolidacaodevendas',
        'filterVenda'        => 'vnd05400_consolidacaodevendas_nrseqvenda',
        'filterCentroCusto'  => 'vnd05400_consolidacaodevendas_cdccusclie',
        'filterProduto'      => 'vnd05400_consolidacaodevendas_cdproduto'
    );
    protected $emailOrig = 'retailmanagement@teknisa.com';
    protected $email = 'alline.martins@teknisa.com';
    
    protected $smtp;
    protected $logConsolidation;
    protected $consolidationFileLogService;
    protected $codReferencia;

    public function __construct(ORM\EntityManager $entityManager, $consolidationFileLogService, $environment )  {
        $this->entityManager = $entityManager;
        $this->smtp = new SMTP($entityManager);
        $this->environment = $environment;
        $this->consolidationFileLogService = $consolidationFileLogService;
        $this->logConsolidation = "";
    }
    
    public function getLogService() {
        return $this->consolidationFileLogService;
    }
    
    public function escreverLog ($data) {
        $this->logConsolidation = $this->logConsolidation.$data."<br>\n"; 
    }
    
    public function inicializaLog ($title) {
        $this->logConsolidation = "<div>".$title."</div>"; 
    }
    
    public function finalizaLog($message, $type){
        $this->logConsolidation = $this->logConsolidation." <div class='linha rodape".$type."'><p>".$message."</p></div>";
    }

    public function setDataSourceName(&$dataSourceName, $dataSource){
        $dataSourceName = $this->dataSourceNames[$dataSource];
    }

    public function setParameterBagColumns(&$parameterBagColumns, $conditions){
        $parameterBagColumns = array_column($conditions, 'columnName');
    }

    public function setParameterBagValues(&$parameterBag, $conditions){
        foreach ($conditions as $condition) {
            $columnName = $condition['columnName'];
            $value      = $condition['value'];
            if ($condition['operator'] !== "MAPPED_LIKE_ALL") {
                if (is_array($value))
                    $parameterBag->set($columnName, implode('_', $value));
                else
                    $parameterBag->set($columnName, $value);
            }
        }
    }

    public function newFilter($request, $dataSourceName){
        $filterCriteria = $request->getFilterCriteria();

        return new Request\Filter(
            new FilterCriteria(
                $dataSourceName, 
                Array(), 
                $filterCriteria->getPage(),
                $filterCriteria->getPageSize()
            ),
            $request->getMethod(),
            $request->getRoutePath(),
            $request->getUserId()
        );
    }
    
    public function getParavendParams($cdfilial, $nrorg){
        $params = array(
            ':CDFILIAL' => $cdfilial,
            ':NRORG'    => $nrorg
        );
        $paramsReturn = $this->entityManager->getConnection()->fetchAll(self::GET_PARAVEND_FILIAL, $params);
        
        $pkCount = (is_array($paramsReturn) ? count($paramsReturn) : 0);
        return ($pkCount != 0) ? $paramsReturn[0] : Array();
    }
    
    public function dateRestriction($filterQuery, $selectQuery){
        $params      = null;
        $response    = null;
        $lojaResults = null;
        $localEstoqL = null;
        $message = null;
        
        $dataRestricao = new DateTime($filterQuery['DTINITESCONS']);
        $dataRestricao = $dataRestricao->format('d/m/Y');
        
        $localEstoqL = ($selectQuery['IDUTILLCESTQ'] == 'S') ? '' : 'T';
        if ($selectQuery['IDUTILALMOX'] != "S"){
        // se não utiliza almoxarifado
            $response = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_CONSOLIDACAO_PENDENTE, $filterQuery)[0];
            $lojaResults = array('CDALMOXARIFE' => 'T', 'CDLOCALESTOQ' => $localEstoqL);
            $message = 'Data de restrição parametrizada '.$dataRestricao.'. Considerando que não utiliza Controle por Almoxarifado (Parâmetros de Estoque)';
        }
        else {
            // se não utiliza localização
            $params = array(
                'CDFILIAL' => $filterQuery['CDFILIAL'],
                'CDLOJA'   => $selectQuery['CDLOJA'],
                'NRORG'    => $filterQuery['NRORG']
            );
            $lojaResults = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_LOJA_ALMOXARIFE, $params);
            
            $pkCount = (is_array($lojaResults) ? count($lojaResults) : 0);
            $lojaResults = ($pkCount == 1) ? $lojaResults[0] : array('CDALMOXARIFE' => '', 'CDLOCALESTOQ' => '');
            
            $filterQuery['CDALMOXARIFE'] = $lojaResults['CDALMOXARIFE'];
            $filterQuery['CDLOCALESTOQ'] = $lojaResults['CDLOCALESTOQ'];
            $filterQuery['UTILOCALESTQ'] = $localEstoqL;
            $response = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_CONSOLIDACAO_PENDENTE_ALMOXARIFADO, $filterQuery)[0];
            
            $lojaResults['CDLOCALESTOQ'] = ($localEstoqL == '') ? $lojaResults['CDLOCALESTOQ'] : 'T';
            $message = 'Data de restrição parametrizada '.$dataRestricao.'. Considerando que utiliza Controle por Almoxarifado '. ($selectQuery['IDUTILLCESTQ'] == 'S') ? 'e utiliza Controle de Local de Estoque' : '' .' (Parâmetros de Estoque)';
        }
        
        return $response = array(
            'status'      => ($response['NRVENDA'] == 0) ? true : false,
            'lojaResults' => $lojaResults,
            'message' => $message
        );
    }
    
    public function getParamEstoque($nrorg) {
        $params = array(
            ':NRORG'    => $nrorg
        );
        
        $paramsReturn = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_PARAMESTOQUE, $params);
        $pkCount = (is_array($paramsReturn) ? count($paramsReturn) : 0);
        return ($pkCount != 0) ? $paramsReturn[0] : Array();
    }
    
    public function consolidationForSaleDate($products, $filterParams, $filialParams, $paramEstoque) {
        $consolidationForSaleDate = array(
            'status'  => false,
            'message' => ''
        );
        
        $dateSale       = array();
        $processingDate = array();
        
        $this->email = ($filterParams['sendEmail'] == '') ? $filterParams['sendEmail'] : 'alline.martins@teknisa.com';
        $this->codReferencia = $filterParams['NRORG'].$filterParams['CDFILIAL'].$filterParams['CDLOJA'];
        
        try {
            // unifica a data das vendas
            $dateSale       = self::unifiesArray($products, 'DTENTRVENDA');
            $this->consolidationFileLogService->saveProcessLog('okSalesUnified', '', $this->codReferencia);
            
            // checa data de processamento pela data da venda, de acordo com a PARAVEND e PARAMESTOQUE
            $processingDate = self::processingDate($dateSale, $filialParams, $paramEstoque);
            $processingDate['status'] = true;
            
            $this->consolidationFileLogService->saveProcessLog('okCheckProcessingDate', '', $this->codReferencia);
            if ($processingDate['status']) {
                $this->consolidationFileLogService->saveProcessLog('okProcessingDateStatusTrue', '', $this->codReferencia);
                
                // verifica implantação ou ajuste de estoque pela data das vendas
                $modifiedStock = self::modifiedStock($dateSale, $filterParams);
                $this->consolidationFileLogService->saveProcessLog('okVerifyModifiedStock', '', $this->codReferencia);
                
                if ($modifiedStock['status']) {
                    $this->consolidationFileLogService->saveProcessLog('okModifiedStockStatusTrue', '', $this->codReferencia);
                    $consolidationForSaleDate['status']  = true;
                    $consolidationForSaleDate['message'] = 'Ok.';
                } else {
                    $this->consolidationFileLogService->saveProcessLog('okModifiedStockStatusFalse', '', $this->codReferencia);
                    $consolidationForSaleDate['message'] = $modifiedStock['message'];
                } 
            }
            else {
                $this->consolidationFileLogService->saveProcessLog('okProcessingDateStatusFalse', $processingDate['message'], $this->codReferencia);
                $consolidationForSaleDate['message'] = $processingDate['message'];
            }
            
            return $consolidationForSaleDate;
        } catch (Exception $e) {
            //$this->saveExceptionLog ('exceptionIntro', $e->getMessage());
             $this->consolidationFileLogService->saveExceptionLog('exceptionIntro', $e, $this->codReferencia);
            throw $e;
        }
    }
    
    public function unifiesArray($products, $index){
        return array_unique(array_column($products, $index));
    }
    
    public function processingDate($dateSale, $filialParams, $paramEstoque) {
        $processingDate = array(
            'status'  => true,
            'message' => 'Ok.'
        );
        $messageProcessing = array(
            'Data_Mai' => 'Data da venda deve ser menor ou igual à data de processamento\data atual de acordo com configuração na Tela Parâmetros de Estoque que não Permite Lançamentos Posteriores à Data de Processamento. ',
            'Data_M'   => 'Data da venda fora do período permitido de acordo com o número de dias para Lançamentos de Saída configurado na Tela Parâmetros de Estoque. ',
            'Data_Men' => 'Data da venda deve ser maior ou igual à data de processamento\data atual de acordo com configuração na Tela Parâmetros de Estoque que não Permite Lançamentos Anteriores à Data de Processamento. '
        );
        
        // Data de Processamento da Venda como DateInterval Object
        $dataProcObj = null;
        //verifica se campo de DTPROCESSA consta vazio, se tiver coloca a data atual
        $filialParams['DTPROCESSA'] = (isset($filialParams['DTPROCESSA']) && !empty($filialParams['DTPROCESSA'])) ? $filialParams['DTPROCESSA'] : date('d/m/Y');
        
        //verifica se campo de NRDIASSAIDPDFIL consta vazio, se tiver coloca 0
        $filialParams['NRDIASSAIDPDFIL'] = (isset($filialParams['NRDIASSAIDPDFIL']) && !empty($filialParams['NRDIASSAIDPDFIL'])) ? $filialParams['NRDIASSAIDPDFIL'] : 0;
        
        //verifica se campo de NRDIASSAIDFIL consta vazio, se tiver coloca 0
        $filialParams['NRDIASSAIDFIL'] = (isset($filialParams['NRDIASSAIDFIL']) && !empty($filialParams['NRDIASSAIDFIL'])) ? $filialParams['NRDIASSAIDFIL'] : 0;
        
        //verifica se campo de NRDIASSAIDPD consta vazio, se tiver coloca 0
        $paramEstoque['NRDIASSAIDPD'] = (isset($paramEstoque['NRDIASSAIDPD']) && !empty($paramEstoque['NRDIASSAIDPD'])) ? $paramEstoque['NRDIASSAIDPD'] : 0;
        
        //verifica se campo de NRDIASSAID consta vazio, se tiver coloca 0
        $paramEstoque['NRDIASSAID'] = (isset($paramEstoque['NRDIASSAID']) && !empty($paramEstoque['NRDIASSAID'])) ? $paramEstoque['NRDIASSAID'] : 0;
        
        // Data de Processamento da Venda como Integer (nativo do PHP)
        $dataProcInt = Utilities\Util::formatDateToInteger($filialParams['DTPROCESSA']);
        // Data de Processamento fixa
        $DTPROCESSA = Utilities\Util::formatDateToInteger($filialParams['DTPROCESSA']);
        foreach($dateSale as $sale) {
            $sale = Utilities\Util::formatDateToInteger($sale);
            if ($sale > $DTPROCESSA) {
                $dataProcObj = DateTime::createFromFormat('d/m/Y', $filialParams['DTPROCESSA']);
                if ($filialParams['IDLANCDTMAIFIL'] == 'N') {
                    //condicao que NÃO permite Lançamentos Posteriores à Data de Processamento
                    if ($paramEstoque['IDLANCDTMAI'] == 'N') {
                        //condicao que NÃO Permite Lançamentos Posteriores à Data de Processamento
                        $processingDate['status']  = false;
                        $message =  'Data de processamento considerada: '.(new \DateTime($dataProcObj))->format('d/m/Y');
                        $processingDate['message'] = $messageProcessing['Data_Mai'].$message;
                        break;
                    }
                    else {
                        //calcula data de acordo com o numero Dias para Lançamentos de Saída NRDIASSAIDPD
                        //condicao que Permite Lançamentos Posteriores à Data de Processamento
                        $dataProcObj->add(new \DateInterval('P'. strval($paramEstoque['NRDIASSAIDPD']) .'D'));
                        $dataProcInt = Utilities\Util::formatDateToInteger($dataProcObj->format('d/m/Y'));
                        if ($sale > $dataProcInt) {
                            $processingDate['status']  = false;
                            $message = 'Data base '. $filialParams['DTPROCESSA'].' mais '.$paramEstoque['NRDIASSAIDPD'].' dias. Data de processamento considerada: '.(new \DateTime($dataProcInt))->format('d/m/Y');
                            $processingDate['message'] = $messageProcessing['Data_M'] . 'Data da Venda: '.(new \DateTime($sale))->format('d/m/Y').' maior que Data de Processamento! '.$message;
                            break;
                        }
                    }
                }
                else{
                    //calcula data de acordo com o numero Dias para Lançamentos de Saída NRDIASSAIDPDFIL
                    //condicao que Permite Lançamentos Posteriores à Data de Processamento
                    $dataProcObj->add(new \DateInterval('P'. strval($filialParams['NRDIASSAIDPDFIL']) .'D'));
                    $dataProcInt = Utilities\Util::formatDateToInteger($dataProcObj->format('d/m/Y'));
                    if ($sale > $dataProcInt) {
                        $processingDate['status']  = false;
                        $message = 'Data base '. $filialParams['DTPROCESSA'].' mais '.$filialParams['NRDIASSAIDPDFIL'].' dias. Data de processamento considerada: '.(new \DateTime($dataProcInt))->format('d/m/Y');
                        $processingDate['message'] = $messageProcessing['Data_M'] . 'Data da Venda: '.(new \DateTime($sale))->format('d/m/Y').' maior que Data de Processamento! '.$message;
                        break;
                    }
                }
            } else if ($sale < $DTPROCESSA) {
                $dataProcObj = DateTime::createFromFormat('d/m/Y', $filialParams['DTPROCESSA']);
                if ($filialParams['IDLANCDTMENFIL'] == 'N'){
                    //condicao que NÃO permite Lançamentos Anteriores à Data de Processamento
                    if ($paramEstoque['IDLANCDTMEN'] == 'N'){
                        //condicao que NÃO Permite Lançamentos Anteriores à Data de Processamento
                        $processingDate['status']  = false;
                        $message =  'Data de processamento considerada: '.(new \DateTime($dataProcInt))->format('d/m/Y');
                        $processingDate['message'] = $messageProcessing['Data_Men'].$message;
                        break;
                    }
                    else {
                        //condicao que Permite Lançamentos Anteriores à Data de Processamento
                        $dataProcObj->sub(new \DateInterval('P'. strval($paramEstoque['NRDIASSAID']) .'D'));
                        $dataProcInt = Utilities\Util::formatDateToInteger($dataProcObj->format('d/m/Y'));
                        if ($sale < $dataProcInt) {
                            $processingDate['status']  = false;
                            $message = 'Data base '. $filialParams['DTPROCESSA'].' menos '.$paramEstoque['NRDIASSAID'].' dias. Data de processamento considerada: '.(new \DateTime($dataProcInt))->format('d/m/Y');
                            $processingDate['message'] = $messageProcessing['Data_M'] . 'Data da Venda: '.(new \DateTime($sale))->format('d/m/Y').' menor que Data de Processamento! '.$message;
                            break;
                        }
                    } 
                }
                else {
                    //calcula data de acordo com o numero Dias para Lançamentos de Saída NRDIASSAIDFIL
                    //condicao que Permite Lançamentos Anteriores à Data de Processamento
                    $dataProcObj->sub(new \DateInterval('P'. strval($filialParams['NRDIASSAIDFIL']) .'D'));
                    $dataProcInt = Utilities\Util::formatDateToInteger($dataProcObj->format('d/m/Y'));
                    if ($sale < $dataProcInt) {
                        $processingDate['status']  = false;
                        $message = 'Data base '. $filialParams['DTPROCESSA'].' menos '.$filialParams['NRDIASSAIDFIL'].' dias. Data de processamento considerada: '.(new \DateTime($dataProcInt))->format('d/m/Y');
                        $processingDate['message'] = $messageProcessing['Data_M'] . 'Data da Venda: '.(new \DateTime($sale))->format('d/m/Y').' menor que Data de Processamento! '.$message;
                        break;
                    }
                }
            }
        }
        return $processingDate;
    }
    
    public function modifiedStock($dateSale, $filterParams) {
        $modifiedStock = array(
            'status'  => true,
            'message' => 'Ok.'
        );
        
        if(is_array($dateSale)){
            if(count($dateSale) > 1){
                $dateSaleI = $dateSale[0];
                foreach($dateSale as $key => $value){
                    if($key != 0){
                        $dateSaleF = $value;
                    }
                }
            }else{
                $dateSaleI = $dateSale[0];
                $dateSaleF = $dateSale[0];
            }
        }else{
            $dateSaleI = $dateSale;
            $dateSaleF = $dateSale;
        }
        
        try {    
            $params = array(
                ':CDFILIAL'   => $filterParams['CDFILIAL'],
                ':DTLANCESTQI' => $dateSaleI,
                ':DTLANCESTQF' => $dateSaleF,
                ':NRORG'      => $filterParams['NRORG']
            );
            
            $stock = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_MODIFIEDSTOCK, $params);
            $this->consolidationFileLogService->saveProcessLog('okModifiedHasBeenFound','', $this->codReferencia);
            $pkCount = (is_array($stock) ? count($stock) : 0);
            if ($pkCount > 0) {
                $this->consolidationFileLogService->saveProcessLog('okStockHigherThanZeroTrue','', $this->codReferencia);
                $dtLanc = $stock[0]['DTLANCESTQ'];
                $modifiedStock = array (
                    'status'  => false,
                    'message' => "Operação bloqueada: Implantação ou Ajuste (último lançamento do dia) já realizada para a data ".$dtLanc."."
                );
            } else {
                $this->consolidationFileLogService->saveProcessLog('okStockHigherThanZeroFalse','', $this->codReferencia);
                $modifiedStock = array (
                    'status'  => true,
                    'message' => 'Ok.'
                );      
            }
        } catch (\Exception $e) {
            //$this->saveExceptionLog ('exceptionIntro', $e->getMessage());
            $this->consolidationFileLogService->saveExceptionLog('exceptionIntro', $e, $this->codReferencia);
            throw $e;
        }
        
        return $modifiedStock;
    }
    
    public function groupConsolidation($filterParams, $filialParams, $paramEstoque, $lojaParams, $products = array()){
        $groupConsolidation = array(
            'errIt' => array()
        );
        
        $params          = array();
        $baixaEstoque    = array();
        $observacoes     = array();
        $paramsControl   = array();
        $perDate         = array();
        $notesPerDate    = null;
        
        $this->inicializaLog('Group Consolidation');
        date_default_timezone_set('America/Sao_Paulo');
        $this->escreverLog("<b>Data: ".date("d/m/Y H:i:s")."</b>");
        
        $this->consolidationFileLogService->saveProcessLog('okGroupingConsolidation', '', $this->codReferencia);
        try {
            // carrega CDALMOXARIFE e CDLOCALESTOQ 
            $pkCount = (is_array($lojaParams) ? count($lojaParams) : 0);
            if ($pkCount == 0) {
                self::getAlmoLocalEstoq($lojaParams, $filterParams, $filialParams);
                $this->consolidationFileLogService->saveProcessLog('okLoadedAlmoLocalEstoq', '', $this->codReferencia);
            } else {
                $this->consolidationFileLogService->saveProcessLog('okNotLoadedAlmoLocalEstoq', '', $this->codReferencia);
            } 
            
            // seta parâmetros da BAIXAESTOQUE e OBSERVACOES
            self::setParamsFilter($params, $filterParams);
            
            $paramsStr =        "CDFILIAL: "      .$params[':CDFILIAL']        ."<br>".   
                                "CDLOJA: "        .$params[':CDLOJA']          ."<br>".     
                                "NRORG: "         .$params[':NRORG']           ."<br>".      
                                "NRSEQVENDA: "    .$params[':NRSEQVENDA']      ."<br>". 
                                "TODASVND: "      .$params[':TODASVND']        ."<br>".   
                                "DTINICIAL: "     .$params[':DTINICIAL']       ."<br>".  
                                "DTFINAL: "       .$params[':DTFINAL']         ."<br>".    
                                "CDCAIXA: "       .$params[':CDCAIXA']         ."<br>".    
                                "TCDCAIXA: "      .$params[':TCDCAIXA']        ."<br>".   
                                "CDCLIENTE: "     .$params[':CDCLIENTE']       ."<br>".  
                                "TCDCLIENTE: "    .$params[':TCDCLIENTE']      ."<br>". 
                                "CDCCUSCLIE: "    .$params[':CDCCUSCLIE']      ."<br>". 
                                "TCDCCUSCLIE: "   .$params[':TCDCCUSCLIE']     ."<br>".
                                "CDPRODUTO: "     .$params[':CDPRODUTO']       ."<br>".  
                                "TCDPRODUTO: "    .$params[':TCDPRODUTO']      ."<br>". 
                                "CDTIPOCONS: "    .$params[':CDTIPOCONS']      ."<br>". 
                                "TCDTIPOCONS: "   .$params[':TCDTIPOCONS']     ."<br>";
            
            $this->escreverLog( $paramsStr );
            $this->consolidationFileLogService->saveProcessLog( 'okProductFilter', $paramsStr , $this->codReferencia);
            // carrega BAIXAESTOQUE de acordo com os produtos da tela
            // Itens da Venda a serem consolidados.
            $baixaEstoque = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_BAIXAESTOQUE, $params);
            $this->escreverLog("<h3>Aqui temos os produtos que serao consolidados:</h3>");
            $this->consolidationFileLogService->saveProcessLog( 'okSaleItemsToBeConsolidated' , '', $this->codReferencia);
            
            foreach ($baixaEstoque as $item) {
                $this->escreverLog($item["CDPRODUTO"].' - '.$item["NMPRODUTO"]);
                $this->consolidationFileLogService->saveProcessLog( 'okItemsDefaultInfo',
                    "CDPRODUTO - " . $item["CDPRODUTO"] .
                    "NMPRODUTO - " . $item["NMPRODUTO"] ,'', $this->codReferencia
                );
            }
            $this->escreverLog("");
            
            // Muda o formato do vetor ->
            // <>Data da Venda<> => <>Item da Venda<> 
            $baixaEstoque = self::mountConsolidationPiece($baixaEstoque, 'DTENTRVENDA');
            // carrega OBSERVACOES - caso IDATUEST == 2 -> não existe observação e composição para os produtos vendidos 
            if ($filialParams['IDATUEST'] == '1') {
                $this->consolidationFileLogService->saveProcessLog( 'okDoExistsObsOrCompProducts', '', $this->codReferencia );
                
                $observacoes = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_OBSERVACOES, $params);
                $this->consolidationFileLogService->saveProcessLog( 'okGetProductsObservations', '', $this->codReferencia );
                $observacoes = self::mountConsolidationPiece($observacoes, 'DTENTRVENDA');
            } else {
                $this->consolidationFileLogService->saveProcessLog( 'okDontExistsObsOrCompProducts', '', $this->codReferencia );
            }
            
            // bloquear consolidação de vendas para produtos inativos ou não cadastrados para a filial
            if ($filialParams['IDBLOQCONSOLIDA'] == 'S') {
                $this->consolidationFileLogService->saveProcessLog( 'okDoBlockConsolidation', '', $this->codReferencia );
                $groupConsolidation['errIt'] = self::VerifyProdFili();
            } else {
                $this->consolidationFileLogService->saveProcessLog( 'okDontBlockConsolidation' , '', $this->codReferencia);
                
            }
            
            // monta array com parâmetros de controle 
            $paramsControl = array(
                'IDATUEST'        => $filialParams['IDATUEST'],
                'CDFILIAL'        => $filterParams['CDFILIAL'],
                'NRORG'           => $filterParams['NRORG'],
                'CDALMOXARIFE'    => $filialParams['CDALMOXARIFE'],
                'CDLOCALESTOQ'    => ($filialParams['CDALMOXARIFE'] == '') ? ' ' : $filialParams['CDLOCALESTOQ'],
                'almoxarifeFL'    => $lojaParams['CDALMOXARIFE'],
                'localestoqFL'    => $lojaParams['CDLOCALESTOQ'],
                'CDTIPCUSTEST'    => ($filialParams['CDTIPCUSTEST'] == '') ? '3' : $filialParams['CDTIPCUSTEST'],
                'DTPROCESSA'      => $filialParams['DTPROCESSA'],
                'CDTIPORETI'      => $filterParams['CDTIPORETI'],
                'IDBLOQCONSOLIDA' => $filialParams['IDBLOQCONSOLIDA'],
                'CDOPERADOR'      => $filterParams['CDOPERADOR']
            );
            
            $groupConsolidation["errIt"] = false;
            $this->finalizaLog('Etapa concluida!', 'Success');
            return [$groupConsolidation, $baixaEstoque, $paramsControl, $observacoes];
        } catch (Exception $e) {
            //$this->saveExceptionLog('exceptionIntro', $e->getMessage());
            $this->consolidationFileLogService->saveExceptionLog('exceptionIntro', $e, $this->codReferencia);
            $this->escreverLog('Erro: '.$e->getMessage());
            $this->finalizaLog('Erro!', 'Erro');
            $this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
            throw $e;
        }
    }
    
    public function agruparProdutosConsolidacao($baixaEstoque, $paramsControl, $observacoes) {
        $groupConsolidation = array(
            'errIt' => array(),
            'data' => array()
        );
        
        foreach ($baixaEstoque as $date => $itens) {
            // verifica se há observação na data da venda
            $this->consolidationFileLogService->saveProcessLog('okCheckForObsInSaleDate','', $this->codReferencia);
            $notesPerDate = array_key_exists($date, $observacoes) ? array_shift($observacoes) : null;
            
            // chama a função que agrupa os itens a serem consolidados por dia
            $this->consolidationFileLogService->saveProcessLog('okGroupItemsByDate','', $this->codReferencia);
            $perDate = self::consolidationPerDate($itens, $notesPerDate, $paramsControl); 
            $pkCountData = (is_array($perDate['data']) ? count($perDate['data']) : 0);
            $pkCountBlock = (is_array($perDate['block']) ? count($perDate['block']) : 0);
            if ($pkCountData  > 0) 
                $groupConsolidation['data'][$date]  = $perDate['data'];
            
            if ($pkCountBlock > 0) 
                $groupConsolidation['block'][$date] = $perDate['block'];
        }
        
        $novoArray = array();
        $novoArray["data"] = [];
        $dias = array_keys($groupConsolidation["data"]);
        $data = $groupConsolidation["data"];
        foreach ($dias as $dia) {
            $consolidacoes = $data[$dia];
            $novoArray["data"][$dia] = array();
            foreach ($consolidacoes as $consolidacao) {
                $index = $this->verificarExistenciaArray($consolidacao, $novoArray["data"][$dia]);
                if ($index >= 0) {
                    $novoArray["data"][$dia][$index]["QTLANCTOEST"] = $novoArray["data"][$dia][$index]["QTLANCTOEST"] + $consolidacao["QTLANCTOEST"];
                    $novoArray["data"][$dia][$index]["VRTOTLANCTO"] = $novoArray["data"][$dia][$index]["VRTOTLANCTO"] + $consolidacao["VRTOTLANCTO"];
                    $novoArray["data"][$dia][$index]["QTTOTLANCTO"] = $novoArray["data"][$dia][$index]["QTTOTLANCTO"] + $consolidacao["QTTOTLANCTO"];
                } else {
                    $novoArray["data"][$dia][] = $consolidacao;
                }
            }
        }
        
        if(!empty($novoArray["data"])){
            foreach($novoArray["data"] as $key => $data){
                $this->consolidationFileLogService->saveProcessLog('okGroupItemsByDateVlrGrp',$key, $this->codReferencia);
                if(!empty($data)){
                    foreach($data as $k => $item){
                        $valueItem = $item["CDPRODUTO"]." - ".$item["NMPRODUTO"]." - QTTOTLANCTO: ".$item["QTTOTLANCTO"]." - QTLANCTOEST: ".$item["QTLANCTOEST"];
                        $this->consolidationFileLogService->saveProcessLog('okItemObservationsInfo', "\nNome do Produto : " .$valueItem, $this->codReferencia);
                        $this->escreverLog("<li>".$valueItem."</li>");
                    }
                }
            }
        }
        
        return $novoArray;
    }
    
    public function verificarExistenciaArray($consolidacao, $array) {
        
        $retorno = -1;
        
        for ($i = 0; $i < sizeof($array); $i++) {
            if (($array[$i]["CDFILIAL"] == $consolidacao["CDFILIAL"])    &&
                ($array[$i]["CDPRODUTO"] == $consolidacao["CDPRODUTO"])  && 
                ($array[$i]["DTLANCMOVI"] == $consolidacao["DTLANCMOVI"]) && 
                ($array[$i]["CDPRODPAI"] == $consolidacao["CDPRODPAI"])) {
                $retorno = $i;
                break;
            }
        }
        
        return $retorno;
    }
    
    public function getAlmoLocalEstoq(&$lojaParams, $filterParams, $filialParams){
    // verifica registro de CDALMOXARIFADO e CDLOCALESTOQ
        $this->escreverLog('Verificando registro de CDALMOXARIFADO e CDLOCALESTOQ ');
        $paramsFilter = array();
        $localEstoqL  = '';
        
        try {
            $localEstoqL = ($filialParams['IDUTILLCESTQ'] == "S") ? '' : 'T';
            
            if ($filialParams['IDUTILALMOX'] != 'S'){
                // se não utiliza almoxarifado
                $lojaParams = array(
                    'CDALMOXARIFE' => 'T', 
                    'CDLOCALESTOQ' => $localEstoqL
                );
            }
            else {
                $paramsFilter = array(
                    'CDFILIAL' => $filterParams['CDFILIAL'],
                    'CDLOJA'   => $filterParams['CDLOJA'],
                    'NRORG'    => $filterParams['NRORG']
                );
                $lojaParams = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_LOJA_ALMOXARIFE, $paramsFilter);
                $pkCount = (is_array($lojaParams) ? count($lojaParams) : 0);
                $lojaParams = ($pkCount == 1) ? $lojaParams[0] : array('CDALMOXARIFE' => '', 'CDLOCALESTOQ' => '');
                
                $lojaParams['CDLOCALESTOQ'] = ($localEstoqL == '') ? $lojaParams['CDLOCALESTOQ'] : 'T';
            }
        } catch (\Exception $e) {
            $this->escreverLog('Erro: '.$e->getMessage());
            $this->finalizaLog('Erro!', 'Erro');
            $this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
            throw $e;
        }
    }
    
    public function setParamsFilter(&$params, $filterParams){
        $params = array(
            ':CDFILIAL'    => $filterParams['CDFILIAL'],
            ':CDLOJA'      => $filterParams['CDLOJA'],
            ':NRORG'       => strVal($filterParams['NRORG']), // banco com dados inconsistentes, query não filtra tudo por NRORG
            ':NRSEQVENDA'  => is_array($filterParams['NRSEQVENDA']) ? implode('_', $filterParams['NRSEQVENDA']) : '',
            ':TODASVND'    => $filterParams['TODASVND'],
            ':DTINICIAL'   => $filterParams['DTINICIAL'],
            ':DTFINAL'     => $filterParams['DTFINAL'],
            ':CDCAIXA'     => is_array($filterParams['CDCAIXA']) ? implode('_', $filterParams['CDCAIXA']) : '',
            ':TCDCAIXA'    => $filterParams['TCDCAIXA'],
            ':CDCLIENTE'   => is_array($filterParams['CDCLIENTE']) ? implode('_', $filterParams['CDCLIENTE']) : '',
            ':TCDCLIENTE'  => $filterParams['TCDCLIENTE'],
            ':CDCCUSCLIE'  => is_array($filterParams['CDCCUSCLIE']) ? implode('_', $filterParams['CDCCUSCLIE']) : '',
            ':TCDCCUSCLIE' => $filterParams['TCDCCUSCLIE'],
            ':CDPRODUTO'   => is_array($filterParams['CDPRODUTO']) ? implode('_', $filterParams['CDPRODUTO']) : '',
            ':TCDPRODUTO'  => $filterParams['TCDPRODUTO'],
            ':CDTIPOCONS'  => is_array($filterParams['CDTIPOCONS']) ? implode('_', $filterParams['CDTIPOCONS']) : '',
            ':TCDTIPOCONS' => $filterParams['TCDTIPOCONS']
        );
    }
    
    public function mountConsolidationPiece($pieces, $index) {
    // monta array ordenado pelo $index
        $arrFormat = array();
        
        $arrFormat = array_fill_keys(self::unifiesArray($pieces, $index), array());
        foreach($pieces as $piece){
            array_push($arrFormat[$piece[$index]], $piece);
        }
        
        return $arrFormat;
    }
    
    public function consolidationPerDate($itens, $notesPerDate, $paramsControl) {
        $notesPerProduct = null;
        $cdproduto       = '';
        $cdprodpai       = '';
        $notes           = null;
        $cdalmoxarife    = '';
        $cdlocalestoq    = '';
        $filterProdFili  = array(
            'CDFILIAL'  => $paramsControl['CDFILIAL'],
            'CDPRODUTO' => '',
            'NRORG'     => $paramsControl['NRORG']
        );
        $resultQuery     = array(
            'already' => false,
            'data'    => ''
        );
        $perProduct      = array();
        
        if ($notesPerDate != null) {
            $notesPerProduct = self::mountConsolidationPiece($notesPerDate, 'CDPRODPAI');
        }
        try {
            foreach($itens as $item) {
                $cdproduto = $item['CDPRODUTO'];
                $cdprodpai = $item['CDPRODPAI'];

                // Verifica se há observação pelo cdproduto da ITEMVENDA
                if (isset($notesPerProduct[$cdprodpai])){
                    $notes = $notesPerProduct[$cdprodpai];
                    unset($notesPerProduct[$cdprodpai]);
                    $this->consolidationFileLogService->saveProcessLog('okItemHasAnObservation', $cdproduto, $this->codReferencia);
                } else {
                    $notes = null;
                    $this->consolidationFileLogService->saveProcessLog('okItemHasntAnObservation', $cdproduto, $this->codReferencia);
                } 
                // almoxarifado para item
                if ($paramsControl['almoxarifeFL'] == ''){
                    $this->consolidationFileLogService->saveProcessLog('okItemDontUseWarehouse', $cdproduto, $this->codReferencia);
                    $filterProdFili['CDPRODUTO'] = $cdproduto;
                    $resultQuery2 = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_PARAM_PRODFILI, $filterProdFili);
                    $pkCount = (is_array($resultQuery2) ? count($resultQuery2) : 0);
                    $resultQuery['data'] = ($pkCount == 0) ? array('CDALMOXARIFE' => '', 'CDLOCALESTOQ' => '') : $resultQuery2[0];
                    if ($resultQuery['data']['CDALMOXARIFE'] == '')
                        $cdalmoxarife = $paramsControl['CDALMOXARIFE'];
                    else 
                        $cdalmoxarife = $resultQuery['data']['CDALMOXARIFE'];
                        
                    $resultQuery['already'] = true;
                } else {
                    if(!($paramsControl['almoxarifeFL'] == 'T')) {
                        $this->consolidationFileLogService->saveProcessLog('okItemDoUseWarehouse', $cdproduto, $this->codReferencia);
                        $cdalmoxarife = $paramsControl['almoxarifeFL']; 
                    } 
                }
                // local de estoque para item
                if ($paramsControl['localestoqFL'] == '') {
                    $this->consolidationFileLogService->saveProcessLog('okItemDontUseStockLocation', $cdproduto, $this->codReferencia);
                    
                    $filterProdFili['CDPRODUTO'] = $cdproduto;
                    
                    if (!$resultQuery['already']) {
                        $this->consolidationFileLogService->saveProcessLog('okGetParamProdFili', $cdproduto, $this->codReferencia);
                        $resultQuery2 = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_PARAM_PRODFILI, $filterProdFili);
                        $pkCount = (is_array($resultQuery2) ? count($resultQuery2) : 0);
                        $resultQuery['data'] = ($pkCount == 0) ? array('CDALMOXARIFE' => '', 'CDLOCALESTOQ' => '') : $resultQuery2[0];
                    }
                    
                    if ($resultQuery['data']['CDLOCALESTOQ'] == '') 
                        $cdlocalestoq = $paramsControl['CDLOCALESTOQ'];
                    else 
                        $cdlocalestoq = ($resultQuery['data']['CDALMOXARIFE'] == '') ? '' : $resultQuery['data']['CDLOCALESTOQ'];
                } else {
                    $this->consolidationFileLogService->saveProcessLog('okItemDoUseStockLocation', $cdproduto, $this->codReferencia);
                    $cdlocalestoq = ($paramsControl['localestoqFL'] == 'T') ? '' : $paramsControl['localestoqFL']; 
                }
                
                // chama a função que agrupa os itens a serem consolidados por produto
                $perProduct = array_merge_recursive($perProduct, self::consolidationPerProduct($item, $notes, $paramsControl, $cdalmoxarife, $cdlocalestoq));
                
                $resultQuery['already'] = false;
                $resultQuery['data']    = array();
            }
            
            return $perProduct;
        } catch (Exception $e) {
            $this->consolidationFileLogService->saveExceptionLog('exceptionIntro', $e, $this->codReferencia);
            $this->escreverLog('Erro: '.$e->getMessage());
            $this->finalizaLog('Erro!', 'Erro');
            $this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
            throw $e;
        }
    }
    
    public function consolidationPerProduct($item, $notes, $paramsControl, $cdalmoxarife, $cdlocalestoq) {
        $recursive  = array();
        $paramRotF  = array();
        $groupItens = array();
        $noteCompo  = null;
        $groupNote  = array();
        $groupBlock = array();
        $groupAll   = array(
            'data'  => array(),
            'block' => array()
        );
        
        $keyNote      = array();
        $keyItem      = array();
        $keyIntersect = array();
        
        if (!empty($notes)) {
            $this->escreverLog("O item <b>".$item["CDPRODUTO"]. " - " .$item["NMPRODUTO"]."</b> possui as seguintes observacoes:");
            
            $this->escreverLog("<ul>");
            foreach ($notes as $note) {
                $this->consolidationFileLogService->saveProcessLog(
                                                    'okItemObservationsInfo', 
                                                    "\nNome do Produto : " .
                                                    $item["CDPRODUTO"] . "\nNome do Produto : " . 
                                                    $item['NMPRODUTO'] . "\nObs: " . 
                                                    $note["IDCONTROLAOBS"] . "\nQTPRODVEND : " . 
                                                    $note["QTPRODVEND"],  $this->codReferencia);
                                                    
                $this->escreverLog("<li>".$note["CDPRODUTO"]." ".$note["NMPRODUTO"]. " Obs: ".$note["IDCONTROLAOBS"]."</li>");
                $this->escreverLog("<li>QTPRODVEND: ".$note["QTPRODVEND"]."</li>");
            }
            $this->escreverLog("</ul>");
        }
        
        try {   
            if ($paramsControl['IDATUEST'] != '1') {
            // item que não existe observação e composição para os produtos vendidos
                $this->consolidationFileLogService->saveProcessLog('okItemDontHaveObsOrComp','', $this->codReferencia);
                array_push($groupItens, $item);
                self::getLancEstValuePerProduct($groupItens, $paramsControl, $cdalmoxarife, $cdlocalestoq);
            } else {
                //chama função recursiva
                $this->consolidationFileLogService->saveProcessLog('okItemDoHaveObsOrComp','', $this->codReferencia);
                $recursive = array(
                    'return'   => array(),
                    'validate' => array($item)
                );
                $paramRotF = array(
                    ':CDFILIAL'  => $paramsControl['CDFILIAL'],
                    ':CDPRODUTO' => '', 
                    ':NRORG'     => $paramsControl['NRORG']
                );
                
                // carrega composição de um item
                $this->consolidationFileLogService->saveProcessLog('okLoadItemComposition','', $this->codReferencia);
                self::getLastProduct($recursive, $paramRotF);
                $groupItens = $recursive['return'];
                
                // calcula valor de lançamento
                $this->consolidationFileLogService->saveProcessLog('okCalcItemVlrLanc','', $this->codReferencia);
                self::getLancEstValuePerProduct($groupItens, $paramsControl, $cdalmoxarife, $cdlocalestoq);
                
                // verifica se existem observações
                if (!empty($notes)) {
                    $this->consolidationFileLogService->saveProcessLog('okItemDontHaveObs','', $this->codReferencia);
                    
                    $groupItens   = array_column($groupItens, null, 'CDPRODUTO');
                    $keyItem      = array_keys($groupItens);
                    $keyNote      = self::unifiesArray($notes, 'CDPRODUTO');
                    $keyIntersect = array_intersect($keyItem, $keyNote); // SEM OBSERVAÇÃO 
                    
                    // caso houver, tira das observações aqueles que estão na composição principal
                 
                    foreach ($keyIntersect as $index) {
                        $keyNote = array();  
                        
                        foreach ($notes as $indexNote => $note) {
                            if ($note['CDPRODUTO'] == $index){
                                array_push($keyNote, $indexNote);
                            }                   
                        }
                        
                        foreach ($keyNote as $key) {
                            // elimina da observação os itens que estão na composição principal
                            $noteCompo = $notes[$key];
                            
                            // trata observações que estão na composição principal
                            if ($noteCompo['IDCONTROLAOBS'] == 'O')
                                $groupItens[$index]['QTLANCTOEST'] += ($groupItens[$index]['QTLANCTOEST']/$groupItens[$index]['QTTOTLANCTO']) * intval($noteCompo['QTPRODVEND']);    
                            else if ($noteCompo['IDCONTROLAOBS'] == 'R') {
                                $groupItens[$index]['QTLANCTOEST'] -= $groupItens[$index]['QTLANCTOEST'] * intval($noteCompo['QTPRODVEND']);
                            }
                            
                            $groupItens[$index]['VRTOTLANCTO'] = $groupItens[$index]['VRUNILANCTO'] * $groupItens[$index]['QTLANCTOEST'];
                        }
                        
                        $newNotes = array();
                        
                        foreach ($notes as $key => $obs) {
                            if (!in_array($key, $keyNote)){
                                array_push($newNotes, $obs);
                            }
                        }
                        
                        $notes = $newNotes;
                    }
                    
                    // se houver observações que não estão na composição principal, são tratadas como itens vendidos separadamente
                    $pkCount = (is_array($notes) ? count($notes) : 0);
                    if ($pkCount != 0) {
                        $this->consolidationFileLogService->saveProcessLog('okSeparatedItems', '', $this->codReferencia);
                        foreach($notes as $note) {
                            $recursive = array(
                                'return'   => array(),
                                'validate' => array($note)
                            );
                            self::getLastProduct($recursive, $paramRotF);
                            $groupNote = $recursive['return'];
                            
                            self::getLancEstValuePerProduct($groupNote, $paramsControl, $cdalmoxarife, $cdlocalestoq);
                        }
                    }
                    
                    // junta composição com observações avulsas  
                    $groupItens = array_merge(array_values($groupItens), $groupNote);
                }
                
                // bloqueia itens não cadastrados por filial ou inativos
                if ($paramsControl['IDBLOQCONSOLIDA'] == 'S') {
                    $groupBlock = self::VerifyItens($groupItens);
                }
            }
            
            $groupAll['data']  = $groupItens;
            $groupAll['block'] = $groupBlock;
            
            $this->escreverLog("");
            
            return $groupAll;
        } catch (Exception $e) {
            throw $e;
        }
    }
    
    public function getLastProduct(&$returnItem, $paramRotF) {
    // função recursiva que retorna os filhos de um item (caso seja composição)
        $formatItem  = ''; 
        $toValidate  = array();
        $composition = array(
            'unique' => false,
            'sons'   => array()
        );
        
        $this->consolidationFileLogService->saveProcessLog('okReturnItemsSons', '', $this->codReferencia);
                        
        try {
            // recebe item a ser avaliado 
            $toValidate = array_shift($returnItem['validate']);
            $paramRotF[':CDPRODUTO'] = $toValidate['CDPRODUTO'];
            // $this->escreverLog("Procurando composição de: ".$paramRotF[':CDPRODUTO']);
            // procura por composição
            $this->consolidationFileLogService->saveProcessLog('okSearchComposition', "\nCDPRODUTO : ". $paramRotF[':CDPRODUTO'], $this->codReferencia);
            self::getComposition($paramRotF, $composition);
            if ($composition['unique'])
                array_push($returnItem['return'], $toValidate);
            else {
                foreach ($composition['sons'] as &$sons) {
                    $sons['DTENTRVENDA'] = $toValidate['DTENTRVENDA'];
                    $sons['CDPRODPAI']   = $toValidate['CDPRODPAI'];
                    $sons['QTPRODVEND']  = strval(floatval($sons['QTPRODVEND']) * floatval($toValidate['QTPRODVEND']));
                    // $this->escreverLog("QTPRODVEND: ".$sons['QTPRODVEND']);
                    
                }
                $returnItem['validate'] = array_merge($returnItem['validate'], $composition['sons']);
            }
            // caso houver itens a serem verificados, a função é retomada
            $pkCount = (is_array($returnItem['validate']) ? count($returnItem['validate']) : 0);
            if ($pkCount != 0)
                $this->getLastProduct($returnItem, $paramRotF);
        } catch (Exception $e) {
            throw $e;
        }
    }
    
    public function getComposition($paramRotF, &$composition){
    // procura composição nas tabelas ROTFABLOCAL e ROTFABRICA
        $returnRotFL = array();
        $returnRotF  = array();
        try {
            // select na ROTFABLOCAL
            $this->consolidationFileLogService->saveProcessLog('okSelectROTFABLOCAL', "\nCDPRODUTO : ". $paramRotF[':CDPRODUTO'], $this->codReferencia);
            $returnRotFL = $this->entityManager->getConnection()->fetchAll(self::GET_ROTFABLOCAL_ITEMS, $paramRotF);
            $pkCount = (is_array($returnRotFL) ? count($returnRotFL) : 0);
            if ($pkCount != 0){
                $composition['unique'] = false;
                $composition['sons']   = $returnRotFL;
            }
            else {
                // select na ROTFRABICA 
                $this->consolidationFileLogService->saveProcessLog('okSelectROTFABLOCAL2', "\nCDPRODUTO : ". $paramRotF[':CDPRODUTO'], $this->codReferencia);
                $returnRotF = $this->entityManager->getConnection()->fetchAll(self::GET_ROTFABRICA_ITEMS, $paramRotF);
                $pkCount = (is_array($returnRotF) ? count($returnRotF) : 0);
                if ($pkCount != 0) {
                    $composition['unique'] = false;
                    $composition['sons']   = $returnRotF;
                }
                else 
                    $composition['unique'] = true;
            }
        } catch (Exception $e) {
            throw $e;
        }
    }
    
    public function getLancEstValuePerProduct(&$arrItens, $paramsControl, $cdalmoxarife, $cdlocalestoq){
    // executa função de valor de lançamento e calcula o valor para cada item
        $filterFunction = array();
        
        $vrfatoconv  = 1;
        $vrUniLancto = 1;
        $qtLanctoEst = 1;
        $vrTotLancto = 1;
        $vrLanctoEst = 1;

        try {
            $filterFunction = array(
                ':P_CDFILIAL'     => $paramsControl['CDFILIAL'],
                ':P_CDPRODUTO'    => '',
                ':P_NRTIPOCUSTO'  => $paramsControl['CDTIPCUSTEST'],
                ':P_DTREFERENCIA' => '',
                ':P_DTATUAL'      => $paramsControl['DTPROCESSA'],
                ':P_ALMOXARIFE'   => ($cdalmoxarife == '') ? ' ' : $cdalmoxarife,
                ':P_LOCALESTOQ'   => ($cdlocalestoq == '') ? ' ' : $cdlocalestoq,
                ':P_NRLOTEESTQ'   => ' ',
                ':P_NRSUBLOTE'    => ' '
            );
            
            foreach($arrItens as &$item){
                $filterFunction[':P_CDPRODUTO']    = $item['CDPRODUTO'];
                $filterFunction[':P_DTREFERENCIA'] = $item['DTENTRVENDA'];
                
                $this->consolidationFileLogService->saveProcessLog('okProductFilter', $item['CDPRODUTO'], $this->codReferencia);
                $vrUniLancto = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_CUSTO_PRODUTO_LCEST, $filterFunction);
                $valorReferencia = is_numeric($vrUniLancto[0]['VRUNILANCTO']) ? 0 : $vrUniLancto[0]['VRUNILANCTO'];
                $pkCount = (is_array($vrUniLancto) ? count($vrUniLancto) : 0);
                $vrUniLancto = ($pkCount > 0) ? floatval(Utilities\Util::RoundDec($vrUniLancto[0]['VRUNILANCTO'], 5)) : 0;
                
                $filterVrfatoconv = array(':CDPRODUTO' => $item['CDPRODUTO']);
            
                $this->consolidationFileLogService->saveProcessLog('okConversionFactor', $item['CDPRODUTO'], $this->codReferencia);
                $resultVrfatoconv  = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_VRFATOCONV, $filterVrfatoconv);
                
                $vrfatoconv = floatval($resultVrfatoconv[0]['VRFATOCONV']);
                
                if($vrfatoconv == 0) {
                    $vrfatoconv = 1;
                }
                
                $qtLanctoEst = $vrfatoconv * floatval($item['QTPRODVEND']); 
                $vrTotLancto = $vrUniLancto * $qtLanctoEst;
                $vrLanctoEst = $vrUniLancto / $vrfatoconv;
                // formata itens para consolidação
                self::formatConsolidation($item, $paramsControl, $qtLanctoEst, $vrLanctoEst, $vrUniLancto, $vrTotLancto, $cdalmoxarife, $cdlocalestoq);
            }
        } catch (Exception $e) {
            throw $e;
        }
    }
    
    public function formatConsolidation(&$item, $paramsControl, $qtLanctoEst, $vrLanctoEst, $vrUniLancto, $vrTotLancto, $cdalmoxarife, $cdlocalestoq){
    // formato utilizado para inserção dos campos
        $item = array(
            'CDFILIAL'       => $paramsControl['CDFILIAL'],
            'NRLANCESTQ'     => '', // adicionado posteriormente 
            'NRSEQUITEM'     => '', // adicionado posteriormente 
            'CDPRODUTO'      => $item['CDPRODUTO'],
            'CDPRODPAI'      => $item['CDPRODPAI'],
            'NMPRODUTO'      => $item['NMPRODUTO'], // apenas para controle
            'QTTOTLANCTO'    => $item['QTPRODVEND'],
            'QTLANCTOEST'    => $qtLanctoEst,
            'DTLANCMOVI'     => $item['DTENTRVENDA'],
            'VRLANCTOEST'    => $vrLanctoEst,
            'VRLANCTOBRUT'   => $vrLanctoEst,
            'VRUNILANCTO'    => $vrUniLancto,
            'VRTOTLANCTO'    => $vrTotLancto,
            'CDPRODMOVI'     => $item['CDPRODESTO'],
            'IDEVOLESTOQ'    => 'S',
            'CDALMOXARIFE'   => ($cdalmoxarife == '') ? ' ' : $cdalmoxarife,
            'CDLOCALESTOQ'   => ($cdlocalestoq == '') ? ' ' : $cdlocalestoq,
            'NRLOTEESTQ'     => ' ',
            'NRSUBLOTE'      => ' ',
            'CDFILIMOVI'     => $paramsControl['CDFILIAL'], // verificar
            'IDTIPOMOVI'     => '3',
            'IDORDEMCLC'     => '9',
            'CDCENTCUST'     => '', // adicionado posteriormente 
            'QTDEVOLESTOQ'   => 0,
            'CDTIPORETI'     => $paramsControl['CDTIPORETI'],
            'NRORG'          => $paramsControl['NRORG'],
            'NRORGINCLUSAO'  => $paramsControl['NRORG'],
            'CDOPERINCLUSAO' => $paramsControl['CDOPERADOR']        
        );
    }
    
    public function VerifyProdFili(){
    // esta função ainda será transcrita para o PHP
        return true;
    }
    
    public function VerifyItens($arrItens){
    // esta função ainda será transcrita para o PHP
        return true;
    }
    
    public function insertConsolidation($data, $gridItens, $filterParams, $filialParams){
    // salva consolidação
        $insertConsolidation = array(
            'status'    => self::SUCCESS_C,
            'commited'  => true,
            'irregData' => array(),
            'dateBlock' => ''
        );
        
        $descEstq   = '';
        $filterArr  = array(); // utilizado para alguns processos
        $cdcentcust = '';
            $this->inicializaLog('Insert Consolidation');
        
        try {
            $descEstq      = 'Retirada referente à venda '.$filterParams['CDLOJA'].' - '.$filterParams['CDOPERADOR'];
            // $cdcentcust recebe CDCENTCUST por CDSETOR
            $filterArr = array(
                'CDSETOR' => $filterParams['CDSETOR'],
                'NRORG'   => $filterParams['NRORG']
            );
            $cdcentcust = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_CDCENTCUST, $filterArr);
            $pkCount = (is_array($cdcentcust) ? count($cdcentcust) : 0);
            $cdcentcust = ($pkCount > 0) ? $cdcentcust[0]['CDCENTCUST'] : '';
            // agrupa itens do grid principal por data
            $this->escreverLog('Agrupando itens no grid principal por data...');
            $gridItens = self::mountConsolidationPiece($gridItens, 'DTENTRVENDA');
            $this->escreverLog('Eba! Tudo certo no agrupamento! xD <br>');
            
            // chama função para salvar consolidação
            $this->escreverLog('Iniciando o salvamento da consolidacao...');
            foreach ($data as $date => $itens) {
                
               /*
                IDPERCONVESTNEG => PERMITE CONSOLIDAR VENDAS QUE GERAM VALORES NEGATIVOSEM ESTOQUE
                IDVISUPRODNEG => PERMITE VISUALIZAR OS PRODUTOS QUE FICARÃO IRREGULARES NO ESTOQUE
               */
             
                if(!isset($gridItens[$date])){
                    $gridItens[$date] = array();
                }
                self::insertPerDate($insertConsolidation, $itens, $gridItens[$date], $filterParams, $filialParams, $date, $descEstq, $cdcentcust);    
                
                // verifica se a consolidação prossegue ou não 
                if (!$insertConsolidation['commited']){
                    $insertConsolidation['dateBlock'] = $date;
                    break;
                } else{
                    $this->finalizaLog('Vendas consolidadas!', 'Success');
                }
            }
            
            //envia o arquivo de LOG sobre a consolidacao ... criar tratativas caso nao ache informacoes
            //$logGravado = nl2br(utf8_decode(file_get_contents('/tmp/RotinaConsolidacaoRetail_'.date_create()->format('Ymd').'_'.$this->codReferencia.'_processes.txt')));
            //$this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
            
            return $insertConsolidation;
        } catch (\Exception $e) {
                $this->escreverLog('Erro: '.$e->getMessage());
                $this->finalizaLog('Erro!', 'Erro');
                $this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
           throw $e;
        }
    }
    
    public function insertPerDate(&$insertConsolidation, $itens, $gridItens, $filterParams, $filialParams, $date, $descEstq, $cdcentcust){
    // insere consolidação por data
        $nrLancEstq = '';
        $nrRetiEsto = '';
        $cdproduto  = '';
        $nrSeqItem  = 0;
        $dtFinal    = 0;
        
        $paramsInsert = array();
        $cdprodLote   = array();
        $irregData    = array();
        $irregItens   = array();
        $cancelItens  = array();
        $listProdPai  = array();
        
        try {
        // começa transação com o banco
            self::controlTransaction('begin');
        } catch (\Exception $e) {
            $this->escreverLog('Erro: '.$e->getMessage());
            $this->finalizaLog('Erro!', 'Erro');
            $this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
            throw $e;
        }
        
        try {
            // $nrLancEstq recebe número de lançamento de estoque da NOVOCODIGO
            $nrLancEstq = self::getNewCode('LANCTOEST'.$filterParams['CDFILIAL']);
            // $nrRetiEsto recebe Número de retirada do estoque da NOVOCODIGO
            $nrRetiEsto = self::getNewCode('RETIESTO'.$filterParams['CDFILIAL']);
            
            $this->escreverLog('Numero de Lancamento de Estoque: '.$nrLancEstq);
            $this->escreverLog('Numero de Retirada de Estoque: '.$nrLancEstq);
            
            // salva lançamento de estoque
            $this->escreverLog('');
            $this->escreverLog('Salvando lancamento de estoque...');
            self::formatLanctoEst($paramsInsert, $filterParams, $nrLancEstq, $date, $descEstq);
            $this->entityManager->getConnection()->executeQuery(Utilities\Query::INSERE_LANCTOEST_PDV, $paramsInsert);
            $this->escreverLog('Tudo certo! <br>');
            
            // salva retirada de estoque
            $this->escreverLog('Salvando retirada de estoque...');
            self::formatRetiradaEst($paramsInsert, $filterParams, $nrRetiEsto, $nrLancEstq, $date, $descEstq, $cdcentcust);
            $this->entityManager->getConnection()->executeQuery(Utilities\Query::INSERE_RETIRADAEST, $paramsInsert);
            $this->escreverLog('Tudo certo! <br>');
            
            // verifica produtos que apresentam registro na tabela de lote
            $cdprodLote = self::getCdprodLote($itens, $filterParams);
            $pkCount = (is_array($cdprodLote) ? count($cdprodLote) : 0);
            $cdprodLote = ($pkCount > 0) ? array_column($cdprodLote, null, 'CDPRODUTO') : array();
            //$this->escreverLog('Produtos que apresentam registro na tabela de lote: <br>'.(implode('<br>',array_keys($cdprodLote))));
             
            $this->escreverLog('');
            $this->escreverLog("Salvando produtos na tabela de LOTE e ITLANCTOEST...");
            
            foreach ($itens as &$item) {
                $cdproduto = $item['CDPRODUTO'];
                
                // salva produto na tabela de lote caso não houver
                if (!isset($cdprodLote[$cdproduto])){
                    self::formatLote($paramsInsert, $filterParams, $cdproduto);
                    $this->entityManager->getConnection()->executeQuery(Utilities\Query::INSERT_LOTE_PRODUTO, $paramsInsert);
                    // cdproduto é adicionado no array de checagem após inserção
                    $cdprodLote[$cdproduto] = $cdproduto;
                }
                
                // realiza ultima alteracao nos itens para consolidacao
                $item['NRLANCESTQ'] = $nrLancEstq;
                $item['NRSEQUITEM'] = Utilities\Util::Ajusta(strval(++$nrSeqItem), 4);
                $item['CDCENTCUST'] = $cdcentcust;
                $this->escreverLog($item['CDPRODUTO']." - ".$item['NMPRODUTO']);
                $this->escreverLog('<div><span>Informacoes</span>');
                $this->escreverLog('Cod ALMOX: '.$item['CDALMOXARIFE']);
                $this->escreverLog('Cod Local Est: '.$item['CDLOCALESTOQ']);
                $this->escreverLog('Cod Centro Custo: '.$item['CDCENTCUST']);
                $this->escreverLog('Cod Tipo Retirada: '.$item['CDTIPORETI']);
                $this->escreverLog('QTTOTLANCTO: '.$item['QTTOTLANCTO']);
                $this->escreverLog('QTLANCTOEST: '.$item['QTLANCTOEST']);
                $this->escreverLog('VRLANCTOBRUT: '.$item['VRLANCTOBRUT']);
                $this->escreverLog('VRLANCTOEST: '.$item['VRLANCTOEST']);
                $this->escreverLog('VRTOTLANCTO: '.$item['VRTOTLANCTO']);
                $this->escreverLog('VRUNILANCTO: '.$item['VRUNILANCTO'].'</div>');
                
                if ($item['VRLANCTOBRUT'] < 0 || $item['VRLANCTOEST'] < 0 || $item['VRTOTLANCTO'] < 0 || $item['VRUNILANCTO'] < 0) {
                    $this->escreverLog('<div><span class="erroCritico">CUIDADO! Valor(es) negativo(s)</span>');
                    $this->escreverLog('VRLANCTOBRUT: '.$item['VRLANCTOBRUT']);
                    $this->escreverLog('VRLANCTOEST: '.$item['VRLANCTOEST']);
                    $this->escreverLog('VRTOTLANCTO: '.$item['VRTOTLANCTO']);
                    $this->escreverLog('VRUNILANCTO: '.$item['VRUNILANCTO'].'</div>');
                }
                unset($item['NMPRODUTO']);
                // salva lançamento na ITLANCTOEST
                $this->entityManager->getConnection()->executeQuery(Utilities\Query::INSERT_ITLANCTOEST, $item);
            }
            
            $this->escreverLog("Tudo certo!");
            $this->escreverLog('');

            $dtFinal = (Utilities\Util::formatDateToInteger($date) > Utilities\Util::formatDateToInteger($filialParams['DTPROCESSA']))
                ? $date : $filialParams['DTPROCESSA'];
                
            // executa procedure para atualização de estoque
            self::execProcRecEstoq($filterParams, $nrLancEstq, $date, $dtFinal);
            
            // valida estoque negativo
            self::getIrregStock($irregData, $itens, $filterParams, $nrLancEstq, $date);  
            
            if (!empty($irregData)){
                //IDPERCONVESTNEG => PERMITE CONSOLIDAR VENDAS QUE GERAM VALORES NEGATIVOSEM ESTOQUE
                if ($filialParams['IDPERCONVESTNEG'] == 'S'){
                    // consolidação apenas para item e composição que não apresentar estoque negativo 
                    $this->escreverLog('Consolidando Itens...');
                    $irregItens = array_column($irregData, 'CDPRODUTO');
                    $itens      = self::mountConsolidationPiece($itens, 'CDPRODPAI');
                    
                    // agrupa cancelamento
                    //self::getListIrregProduct($cancelItens, $itens, $irregItens);   
                    
                    
                    //if (count($cancelItens) != count($itens)){
                        // exclui da ITLANCTOEST e adiciona na MOVILOTE os itens a serem cancelados 
                        //$this->escreverLog('excluindo da ITLANCTOEST e adicionando na MOVILOTE os itens a serem cancelados');
                        //$listProdPai = array_keys($cancelItens);
                        
                        if($filialParams['IDPERCONVESTNEG'] == 'S' && $filialParams['IDPERMESTNEG'] == 'N'){
                            self::updateItemLaunch($filterParams, $nrLancEstq);
                        }
                        
                        //pelo que foi explicado nao precisa mais dessa parte
                        //self::delItemLaunch($filterParams, $nrLancEstq, $listProdPai);
                        //elf::addMoviLote($filterParams, $nrLancEstq, $cancelItens, $irregItens);
                        // Linha comentada para consolidar toda a venda sem dar baixa no estoque dos itens irregulares.
                        //self::delItemGrid($gridItens, $listProdPai);
                        
                        // executa (novamente) procedure para atualização de estoque
                        self::execProcRecEstoq($filterParams, $nrLancEstq, $date, $dtFinal);
                        // deleta registro da MOVILOTE 
                        // self::delMoviVend($filterParams, $nrLancEstq);
                        self::mergeIrregData($insertConsolidation, $irregData);
                        
                        
                        $this->escreverLog('Tudo certo!');
                    //}
                   /* else {
                        // cancela consolidação caso conter estoque irregular para todas as vendas da data
                        $this->escreverLog('Estoque irregular! Cancelando Consolidação...  :( ');
                        $this->finalizaLog('As vendas não foram Consolidadas!', 'Warning');
                        self::controlTransaction('rollback');
                        $insertConsolidation['status']    = self::FAIL_DATE_C;                    
                        $insertConsolidation['commited']  = false;  
                        $insertConsolidation['dateBlock'] = $date;  
                    }*/
                }
                else{
                    // rollback na consolidação caso parametrização não aceite estoque irregular, se houver 
                    $this->escreverLog('A parametrizacao nao permite estoque irregular! Cancelando Consolidacao...  :(');
                    $this->finalizaLog('As vendas nao foram Consolidadas!', 'Warning');
                    self::controlTransaction('rollback');
                    
                    // executa (novamente) procedure para atualização de estoque
                    self::execProcRecEstoq($filterParams, $nrLancEstq, $date, $dtFinal);
                    
                    //self::getIrregStock($irregData, $itens, $filterParams, $nrLancEstq, $date);
                    $insertConsolidation['status']    = self::FAIL_C;                    
                    $insertConsolidation['commited']  = false;  
                    $insertConsolidation['irregData'] = $irregData;
                }
            }

            if ($insertConsolidation['commited']){
                // atualiza o lançamento de estoque nos itens da venda
                $this->escreverLog('Atualizando o lancamento de estoque nos itens da venda...');
                foreach ($gridItens as $saleItem) {
                    $this->escreverLog("Item: ".$saleItem["NMPRODUTO"]);
                    self::formatItVenda($paramsInsert, $filterParams, $saleItem, $nrLancEstq);
                    $this->entityManager->getConnection()->executeQuery(Utilities\Query::UPDATE_ITEMVENDA, $paramsInsert);
                }
                $this->escreverLog('Tudo certo!!');

                // executa commit da consolidação por data
                self::controlTransaction('commit');
            }
        } catch (\Exception $e) {
            self::controlTransaction('rollback');
            throw $e;
        }
    }
    
    public function controlTransaction($type){
    // controla transação com banco    
        try {
            switch ($type) {
                case 'begin':
                // inicia transação sem auto-commit
                    $this->entityManager->getConnection()->beginTransaction();
                    break;
                case 'commit':
                // salva transação
                    $this->entityManager->getConnection()->commit();
                    break;
                case 'rollback':
                // descarta transação
                    $this->entityManager->getConnection()->rollBack();
                    break;
            }
        } catch (\Exception $e) {
            $this->escreverLog('Erro: '.$e->getMessage());
            $this->finalizaLog('Erro!', 'Erro');
            $this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
            throw $e;
        }
    }
    
    public function getNewCode($param){
    // executa procedure NOVO_CODIGO
        $procedure = null;
        $nrseq     = '';
        
        try {
            $procedure = new StoredProcedure($this->entityManager->getConnection(), 'NOVO_CODIGO');
            $procedure->addParam(new Param('P_CONTADOR'  , Param::PARAM_INPUT , $param, Param::PARAM_TYPE_STR));
            $procedure->addParam(new Param('P_SEQUENCIAL', Param::PARAM_OUTPUT, null  , Param::PARAM_TYPE_INT, 20));
            $nrseq = $procedure->execute();   
            
            return substr($nrseq['P_SEQUENCIAL'], 10, 10);
        } catch (\Exception $e) {
            $this->escreverLog('Erro: '.$e->getMessage());
            $this->finalizaLog('Erro!', 'Erro');
            $this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
            throw $e;
        }
    }
    
    public function formatLanctoEst(&$paramsInsert, $filterParams, $nrLancEstq, $date, $descEstq){
    // formato utilizado para inserção dos campos no lançamento de estoque
        $paramsInsert = array(
            'CDFILIAL'       => $filterParams['CDFILIAL'],
            'NRLANCESTQ'     => $nrLancEstq,
            'IDTPLANCTO'     => '3',
            'DTLANCESTQ'     => $date,
            'DSLANCESTQ'     => $descEstq,
            'IDALTCUSMED'    => 'N',
            'CDFILIMOVI'     => $filterParams['CDFILIAL'],
            'CDOPERADOR'     => $filterParams['CDOPERADOR'],
            'NRORG'          => $filterParams['NRORG'],
            'NRORGINCLUSAO'  => $filterParams['NRORG'],
            'CDOPERINCLUSAO' => $filterParams['CDOPERADOR']        
        );
    }
    
    public function formatRetiradaEst(&$paramsInsert, $filterParams, $nrRetiEsto, $nrLancEstq, $date, $descEstq, $cdcentcust){
    // formato utilizado para inserção dos campos na retirada de estoque
        $paramsInsert = array(
            'CDFILIAL'       => $filterParams['CDFILIAL'],
            'NRRETEST'       => $nrRetiEsto,
            'DTRETEST'       => $date,
            'DSRETEST'       => $descEstq,
            'CDFILIALLAN'    => $filterParams['CDFILIAL'],
            'NRLANCESTQ'     => $nrLancEstq,
            'CDSETOR'        => $filterParams['CDSETOR'],
            'CDCENTCUST'     => $cdcentcust,
            'NRORG'          => $filterParams['NRORG'],
            'NRORGINCLUSAO'  => $filterParams['NRORG'],
            'CDOPERINCLUSAO' => $filterParams['CDOPERADOR']        
        );
    }
    
    public function getCdprodLote($itens, $filterParams){
    // verifica se há registro para os produtos na tabela de lote
        $paramsFilter = array();
        
        try {
            $paramsFilter = array(
                'CDFILIAL'   => $filterParams['CDFILIAL'],
                'NRLOTEESTQ' => ' ',
                'NRSUBLOTE'  => ' ',
                'NRORG'      => $filterParams['NRORG'],
                'CDPRODUTO'  => implode('_', self::unifiesArray($itens, 'CDPRODUTO'))
            );
            
            return $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_LOTE_PRODUTO, $paramsFilter);
        } catch (\Exception $e) {
            $this->escreverLog('Erro: '.$e->getMessage());
            $this->finalizaLog('Erro!', 'Erro');
            $this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
            throw $e;
        }
    }
    
    public function formatLote(&$paramsInsert, $filterParams, $cdproduto){
    // formato utilizado para inserção dos campos no lote
        $paramsInsert = array(
            'CDFILIAL'       => $filterParams['CDFILIAL'],
            'NRLOTEESTQ'     => ' ',
            'NRSUBLOTE'      => ' ',
            'IDSTATLOTE'     => 'L',
            'CDPRODUTO'      => $cdproduto, 
            'NRORG'          => $filterParams['NRORG'],
            'NRORGINCLUSAO'  => $filterParams['NRORG'],
            'CDOPERINCLUSAO' => $filterParams['CDOPERADOR']
        );
    }
    
    public function execProcRecEstoq($filterParams, $nrLancEstq, $dtInicial, $dtFinal){
    // executa procedure de RecEstoq
        $procedure = null;
        $this->escreverLog('Executando procedure para atualizacao de estoque...');
        try {
            $dtInicial = DateTime::createFromFormat('d/m/Y', $dtInicial);
            $dtFinal   = DateTime::createFromFormat('d/m/Y', $dtFinal);
            
            $procedure = new StoredProcedure($this->entityManager->getConnection(), 'ATUALIZA_ESTOQUE');
            //P_FILIMOVI tem que ser a filial que ta fazendo o lançamento
            //P_FILILANCTO tem q ser a filial logada
            $procedure->addParam(new Param('P_FILIMOVI'  , Param::PARAM_INPUT, $filterParams['CDFILIAL']  , Param::PARAM_TYPE_STR));
            $procedure->addParam(new Param('P_DTINICIAL' , Param::PARAM_INPUT, $dtInicial->format('d/m/Y'), Param::PARAM_TYPE_STR));
            $procedure->addParam(new Param('P_DTFINAL'   , Param::PARAM_INPUT, $dtFinal->format('d/m/Y')  , Param::PARAM_TYPE_STR));
            $procedure->addParam(new Param('P_FILILANCTO', Param::PARAM_INPUT, $this->environment->getCdFilial(), Param::PARAM_TYPE_STR));
            $procedure->addParam(new Param('P_NRLANCTO'  , Param::PARAM_INPUT, $nrLancEstq                , Param::PARAM_TYPE_STR));
            $procedure->execute();
            $this->escreverLog('Executada! =) <br>');
        } catch (\Exception $e) {
            $this->escreverLog('Erro: '.$e->getMessage());
            $this->finalizaLog('Erro!', 'Erro');
            $this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
            throw $e;
        }
    }
    
    public function formatItVenda(&$paramsInsert, $filterParams, $saleItem, $nrLancEstq){
    // formato utilizado para update dos campos na ITEMVENDA
        $paramsInsert = array(
            'CDFILIALEST'  => $filterParams['CDFILIAL'], 
            'NRLANCESTQ'   => $nrLancEstq, 
            'NRSEQUITVEND' => $saleItem['NRSEQUITVEND'],
            'CDFILIAL'     => $saleItem['CDFILIAL'], 
            'CDCAIXA'      => $saleItem['CDCAIXA'], 
            'NRSEQVENDA'   => $saleItem['NRSEQVENDA'],
            'NRORG'        => $filterParams['NRORG'],
            'NRORGULTATU'  => $filterParams['NRORG'],
            'CDOPERULTATU' => $filterParams['CDOPERADOR']
        );
    }
   
    public function getIrregStock(&$irregData, $itens, $filterParams, $nrLancEstq, $date){
    // busca itens com estoque irregular e calcula seu valor negativo
        $this->escreverLog('<b>Buscando produtos irregulares...</b>');
        $paramsFilter = array();
        $irregItens   = array();
        
        $irregProd    = ''; 
        $cdalmoxarife = '';
        $cdlocalestoq = '';
        $qtEstoqDia   = 0;
        $estqArvPro   = 0;
        
        try {
            $paramsFilter = array(
                'CDFILINRLANC' => $filterParams['CDFILIAL'].$nrLancEstq,
                'NRORG'        => $filterParams['NRORG'], 
                'CDFILIAL'     => $filterParams['CDFILIAL']
            ); 
            $irregItens = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_MOVILOTE, $paramsFilter);
            $pkCount = (is_array($irregItens) ? count($irregItens) : 0);
            if ($pkCount > 0){
               // incrementa 1 dia na data da venda para filtro da POSESTDIA 
                $date = DateTime::createFromFormat('d/m/Y', $date);
                $date->add(new \DateInterval('P1D'));
                
                $itens = array_column($itens, null, 'CDPRODMOVI');
                $this->escreverLog('<b>Produto irregular encontrado!</b><br>');
                $this->escreverLog('Produtos irregulares na data: '.$date.'<br>');
                foreach ($irregItens as &$irregular) {
                    $irregProd = $irregular['CDPRODUTO'];

                    $cdalmoxarife = $itens[$irregProd]['CDALMOXARIFE'];
                    $cdlocalestoq = $itens[$irregProd]['CDLOCALESTOQ'];
                    
                    // busca quantidade de estoque pelo dia
                    $paramsFilter = array(
                        'CDFILIAL'     => $filterParams['CDFILIAL'],
                        'CDPRODUTO'    => $irregProd,
                        'CDALMOXARIFE' => $cdalmoxarife,
                        'CDLOCALESTOQ' => $cdlocalestoq,
                        'NRLOTEESTQ'   => ' ',
                        'NRSUBLOTE'    => ' ',
                        'DTPOSIESTQ'   => $date->format('d/m/Y'),
                        'NRORG'        => $filterParams['NRORG'] 
                    );
                    $estqArvPro = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_QTESTOQ_ARVPROD, $paramsFilter);
                    if(!empty($estqArvPro)){
                        if(isset($estqArvPro[0])){
                            $estqArvPro = $estqArvPro[0];
                            $irregular['QTFINAL']   = floatval($estqArvPro['QTESTOQDIA']);
                            $this->escreverLog('<li>'.$irregProd.' -     Quantidade: '.$irregular['QTFINAL'].'</li>');
                            $irregular['CDARVPROD'] = $estqArvPro['CDARVPROD'];
                            $irregular['CDFILIAL']  = $filterParams['CDFILIAL'];
                        }
                    }
                }
                $irregData = $irregItens;
            } else {
                $this->escreverLog('Eba! Nenhum produto irregular foi encontrado!');
            }
            
        } catch (\Exception $e) {
            $this->escreverLog('Erro: '.$e->getMessage());
            $this->finalizaLog('Erro!', 'Erro');
            $this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
            throw $e;
        }
    }
    
    public function delMoviVend($filterParams, $nrLancEstq){
    // limpa qualquer rastro da consolidação na tabela de MOVILOTE
        $paramsFilter = array();
        
        try {
            $paramsFilter = array(
                'CDFILINRLANC' => $filterParams['CDFILIAL'].$nrLancEstq,
                'NRORG'        => $filterParams['NRORG']
            );
            
            $this->entityManager->getConnection()->executeQuery(Utilities\Query::DELETE_MOVILOTE, $paramsFilter); 
        } catch (\Exception $e) {
            $this->escreverLog('Erro: '.$e->getMessage());
            $this->finalizaLog('Erro!', 'Erro');
            $this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
            throw $e;
        }
    }

    public function getListIrregProduct(&$cancelItens, $itens, $irregItens){
    // separa lista que será cancelada, indexada pelo cdprodpai
        $cancelItens = $itens;
        
        foreach ($itens as $cdprodpai => $sons) {
            $sons = array_unique(array_column($sons, 'CDPRODMOVI'));
            
            if (empty(array_intersect($sons, $irregItens)))
                unset($cancelItens[$cdprodpai]);
        }
    }

    public function updateItemLaunch($filterParams, $nrLancEstq){
    // update na ITLANCTOEST os itens irregulares e que pertencem à sua composição
        $paramsFilter = array();
        
        try {
            $paramsFilter = array(
                'CDFILIAL' => $filterParams['CDFILIAL'],
                'NRLANCESTQ' => $nrLancEstq
            );
            $this->entityManager->getConnection()->executeQuery(Utilities\Query::UPDATE_ITLANCTOEST, $paramsFilter);
        } catch (\Exception $e) {
            $this->escreverLog('Erro: '.$e->getMessage());
            $this->finalizaLog('Erro!', 'Erro');
            $this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
            throw $e;
        }   
    }
    public function delItemLaunch($filterParams, $nrLancEstq, $listProdPai){
    // exclui da ITLANCTOEST os itens irregulares e que pertencem à sua composição
        $paramsFilter = array();
        
        try {
            $paramsFilter = array(
                'CDFILIAL' => $filterParams['CDFILIAL'],
                'NRLANCESTQ' => $nrLancEstq,
                'CDPRODPAI' => implode('_', $listProdPai),
                'CDFILIMOVI' => $filterParams['CDFILIAL'],
                'NRORG' => $filterParams['NRORG']
            );
            
            $this->entityManager->getConnection()->executeQuery(Utilities\Query::DELETE_ITLACTOEST, $paramsFilter);
        } catch (\Exception $e) {
            $this->escreverLog('Erro: '.$e->getMessage());
            $this->finalizaLog('Erro!', 'Erro');
            $this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
            throw $e;
        }   
    }

    public function addMoviLote($filterParams, $nrLancEstq, $cancelItens, $irregItens){
    // agrupa e adiciona na MOVILOTE itens não irregulares que serão cancelados
        $paramsFilter = array();
        $paramsInsert = array();
        $getMoviLote  = array();

        try {           
            $paramsFilter = array(
                'CDFILINRLANC' => $filterParams['CDFILIAL'].$nrLancEstq,
                'NRORG'        => $filterParams['NRORG'],
                'CDFILIAL'     => $filterParams['CDFILIAL']
            );
            $getMoviLote = $this->entityManager->getConnection()->fetchAll(Utilities\Query::GET_MOVILOTE, $paramsFilter);
            $irregItens  = array_unique(array_merge($irregItens, array_column($getMoviLote, 'CDPRODUTO')));

            foreach ($cancelItens as $cdprodpai) {
                foreach ($cdprodpai as $item) {
                    if (!in_array($item['CDPRODMOVI'], $irregItens)){                        
                        self::formatMoviLote($paramsInsert, $item);

                        $this->entityManager->getConnection()->executeQuery(Utilities\Query::INSERT_MOVILOTE, $paramsInsert); 
                        array_push($irregItens, $item['CDPRODMOVI']);
                    }
                }
            }            
        } catch (\Exception $e) {
            $this->escreverLog('Erro: '.$e->getMessage());
            $this->finalizaLog('Erro!', 'Erro');
            $this->email ? $this->smtp->sendEmailMessage("manage", $this->email, $this->emailOrig, $this->logConsolidation) : 0;
            throw $e;
        }        
    }    

    public function formatMoviLote(&$paramsInsert, $item){
    // formato utilizado para inserção dos campos na MOVILOTE
        $paramsInsert = array(
            'NRMOVILOTE'    => $item['CDFILIAL'].$item['NRLANCESTQ'],
            'CDFILIAL'      => $item['CDFILIMOVI'],
            'CDALMOXARIFE'  => $item['CDALMOXARIFE'],
            'CDLOCALESTOQ'  => $item['CDLOCALESTOQ'],
            'CDPRODUTO'     => $item['CDPRODMOVI'],
            'NRLOTEESTQ'    => $item['NRLOTEESTQ'],
            'NRSUBLOTE'     => $item['NRSUBLOTE'],
            'DTMOVILOTE'    => $item['DTLANCMOVI'],
            'DTMOVILOTEANT' => $item['DTLANCMOVI'],
            'NRORG'         => $item['NRORG']
        );
    }

    public function delItemGrid(&$gridItens, $listProdPai){
    // remove itens que não serão consolidados da lista de update da ITVENDA
        $itens = $gridItens;

        foreach ($itens as $key => $item) {
            if(in_array($item['CDPRODUTO'], $listProdPai))
                unset($gridItens[$key]);
        }
    }

    public  function mergeIrregData(&$insertConsolidation, $irregData){
    // merge entre itens irregulares com tratamento para mesmo CDPRODUTO
        $oldIrregList = array_column($insertConsolidation['irregData'], null, 'CDPRODUTO');
        $newIrregList = array_column($irregData, null, 'CDPRODUTO');

        foreach ($oldIrregList as $cdproduto => &$item) {
            if (isset($newIrregList[$cdproduto])){
                $item = $newIrregList[$cdproduto];
                unset($newIrregList[$cdproduto]);
            }
        }

        $insertConsolidation['irregData'] = array_values(array_merge($oldIrregList, $newIrregList));
    }
    
    public function habilitaPesqProd($params) {
        return $this->entityManager->getConnection()->fetchAll(self::HABILITA_PESQ_PROD, $params);
    }
    
    const GET_PARAVEND_FILIAL = "
       SELECT P.IDCONSUMIOBR,
              P.DTINITESCONS,
              P.CDLOJA,
              P.IDCONSREDZMECF,
              P.IDATUEST,
              P.IDGERANFVENPROD,
              P.IDBLOQCONSOLIDA,
              P.CDTIPORETI,
              P.IDPERCONVESTNEG,
              P.CDSETOR,
              F.CDTIPCUSTEST,
              F.IDUTILALMOX,
              F.IDUTILLCESTQ,
              F.IDCTRLESTQ,
              F.IDPERMESTNEG,
              F.IDMENSESTNEG,
              F.IDLANCDTMENFIL,
              F.NRDIASSAIDFIL,
              F.IDLANCDTMAIFIL,
              F.NRDIASSAIDPDFIL,
              F.CDALMOXARIFE,
              F.CDLOCALESTOQ,
              DECODE(PG.IDUTILDTPROC, 'S', TO_CHAR(F.DTPROCESSA, 'DD/MM/YYYY'), TO_CHAR(SYSDATE, 'DD/MM/YYYY')) AS DTPROCESSA,
              P.IDVISUPRODNEG,
              TR.NMTIPORETI,
              S.NMSETOR
            FROM PARAMGERAL PG
            INNER JOIN PARAVEND P ON (PG.NRORG = P.NRORG)
            INNER JOIN PARAMFILIAL F ON (P.CDFILIAL   = F.CDFILIAL AND P.NRORG      = F.NRORG )
            LEFT JOIN TIPORETI TR ON (TR.CDTIPORETI = P.CDTIPORETI)
            LEFT JOIN SETOR S ON (S.CDSETOR = P.CDSETOR)
            WHERE (P.CDFILIAL = :CDFILIAL )
            AND (P.NRORG      = :NRORG )
    ";
    
    const GET_ROTFABLOCAL_ITEMS = "
        SELECT
          RF.CDSUBPRODLOC AS CDPRODUTO, P.NMPRODUTO, P.CDPRODESTO, NVL(PF.IDCNTRESTOQ, 'S') AS IDCNTRESTOQ, P.VRFATOCONV, RF.QTSUBPRODLOC AS QTPRODVEND
        FROM
          ROTFABLOCAL RF, PRODFILI PF, PRODUTO P, PRODUTO P2
        WHERE (RF.CDFILIAL  = :CDFILIAL      )
          AND (RF.CDPRODUTO = :CDPRODUTO     )
          AND (RF.NRORG     = :NRORG         )
          AND (RF.CDSUBPRODLOC = P.CDPRODUTO )
          AND (RF.NRORG     = P.NRORG        )
          AND (RF.CDFILIAL  = PF.CDFILIAL    )
          AND (RF.NRORG     = PF.NRORG       )
          AND (PF.CDPRODUTO = P.CDPRODESTO   )
          AND (P2.CDPRODUTO = :CDPRODUTO     )
          AND (P2.IDUTILRETPRODACAB     = 'N')
    ";

    const GET_ROTFABRICA_ITEMS = "
        SELECT
          RF.CDSUBPRODUTO AS CDPRODUTO, P.NMPRODUTO, P.CDPRODESTO, NVL(PF.IDCNTRESTOQ, 'S') AS IDCNTRESTOQ, P.VRFATOCONV, RF.QTSUBPRODROT AS QTPRODVEND
        FROM
          ROTFABRICA RF, PRODFILI PF, PRODUTO P, PRODUTO P2
        WHERE (RF.CDPRODUTO = :CDPRODUTO     )
          AND (RF.NRORG     = :NRORG         )
          AND (RF.CDSUBPRODUTO = P.CDPRODUTO )
          AND (RF.NRORG     = P.NRORG        )
          AND (PF.CDFILIAL  = :CDFILIAL      )
          AND (RF.NRORG     = PF.NRORG       )
          AND (PF.CDPRODUTO = P.CDPRODESTO   )
          AND (P2.CDPRODUTO = :CDPRODUTO     )
          AND (P2.IDUTILRETPRODACAB     = 'N')
    ";
    
    const HABILITA_PESQ_PROD = "
        SELECT IDGERANFVENPROD FROM PARAVEND WHERE CDFILIAL = :CDFILIAL AND NRORG = :NRORG
    ";
}